dojo.provide("zero.resource.DataStore");

dojo.require("dojo.data.api.Read");
dojo.require("dojo.data.api.Identity");

// FIXME? Should we also special-case ZRM attribute names such as 'updated'
var minimalZRMStoreObject = {};
// Define ZRMStore specific-properties (not worrying about namespace collisions with
// properties from ZRM Server resources)
minimalZRMStoreObject.isZRMServerItem = true;
minimalZRMStoreObject.dirtyBit = true;
minimalZRMStoreObject.pseudoIdentity = true;

dojo.declare ("zero.resource.DataStore", null, {

    constructor:  function(/* Object */ zrmStoreParameters) {
        // summary: initializer
        // zrmStoreParameters: {contextRoot: String}
        //
        // The value of zrmStoreParameters.contextRoot is the context root of the
        // application (absolute or relative) from which the store's items are
        // fetched and to which they are saved.  The empty string ("") is a
        // legal (relative URI) value.  The value of the 'contextRoot'
        // property is prepended to the value(s) that clients supply in the
        // fetch() 'keywordArgs.query' parameter.  For example, 'contextRoot'
        // might be set to: "http://localhost:8080" and 'keywordArgs.query'
        // set to "/resources/namevaluepairs" with the resulting REST query
        // being "http://localhost:8080/resources/namevaluepairs"
    
        if (!zrmStoreParameters)
            throw new Error ("Missing 'zrmStoreParameters' parameter");

        if (zrmStoreParameters.resourceCollection === undefined)
            throw new Error ("Missing 'zrmStoreParameters.resourceCollection' property");
    
    	// contextRoot is optional
        if (zrmStoreParameters.contextRoot !== undefined)
	        this.contextRoot = zrmStoreParameters.contextRoot;
	        
        this.resourceCollection = zrmStoreParameters.resourceCollection;
        
        this._init(this);
        this._initMetaData(this);
    
        this._features = {'dojo.data.api.Read': true, 'dojo.data.api.Identity': false, 'dojo.data.api.Write': true, 'dojo.data.api.Notification': true};
    
        this._errorHandler = null;
        var d = new Date();
        this._id = d.toUTCString();
        this._fetchCount = 0;
        this._pseudoIdCounter = 0;
        this._baseZRMCollectionURI = "";
        this._resourceNameInCurrentFetch = null;
    
        //this._saveInProgress = false;
    },
 
    metadata: null,

    contextRoot: "", 
    
    resourceCollection: "",
    
    ZRMConstants: {
        "startParam" : "start",
        "countParam" : "count",
        "sortAttribute" : "order_by"
    },                
                    
    // summary:
    //
    // The ZRMStore implements the dojo.data.api.Read API for data that is
    // accessed via a Zero Resource Model Server (i.e., the server is a 'ZRM
    // Server' and the client is a 'ZRMStore').  Specifically, the ZRM Server
    // is accessed through the documented Zero Resource Model REST APIs
    // (http://www.projectzero.org/wiki/bin/view/Documentation/CoreDevelopersGuideResourceModelHttpAPI)
    // The ZRMStore's interactions with the ZRM Server use the (default) JSON
    // format rather than other formats such as such as Atom, Basic XML, and
    // HTML.
    //		 
    // An ZRMStore is populated by issuing a query ("fetch" in
    // dojo.data.api.Read parlance) to the ZRM Server.  The results of the
    // query consist of zero or more rows ("items" in dojo.data.api.Read
    // parlance), each of which contains zero or more attributes.
    //
    //	An ZRMStore "item" is represented internally as an associative array
    //	such that item[x] is "the value of the 'x' attribute for a specific
    //	tuple." ZRMStore instances "aggressively" evaluate client queries:
    //	i.e., they do not perform lazy evaluation of items, but instead
    //	instantiate all items sent by the server into the store.
    //
    // It is important to realize that ZRMStore is completely agnostic as to
    // whether the back-end source of the data is a relational datastore, an
    // Atom feed, or an XML datastore.  The ZRMStore issues only ZRM REST API
    // calls to the ZRM Server.
    //
    // IMPORTANT assumption that's been confirmed with ZRM developers. The ZRM
    // Server's resources are rendered in JSON such that item properties
    // CANNOT be function, array, or object valued.  In other words, item
    // properties are scalar-valued. 
    //
    // IMPORTANT assumption that's been confirmed with ZRM developers. The ZRM
    // resource schema is fixed, and may not be modified by the client.  For
    // example, the client may not use "setValue()" to add an attribute, and
    // "newItem()" may not pass attributes that are not already part of the
    // schema.
    //
    // An ZRMStore instance flushes its contents between queries, and its
    // implementation of the API assumes that clients know that items are
    // implicitly invalidated after a new query is invoked.  This behavior is
    // partly motivated by the fact that the schema of one result set is
    // typically not the same as the schema of another result set.  The strong
    // connection between a given item and its column (attribute) names
    // implies that an item is useless to a client once a subsequent query is
    // invoked on the ZRMStore. Following a second invocation of fetch(),
    // functions such as "isItem()" will return "false" for an item derived
    // from a previous fetch() invocation.  The behavior of "getValue()" etc
    // on a given item is undefined if the function is invoked after a
    // subsquent fetch() flushes the ZRMStore's contents.

    fetch: function(/* Object */ keywordArgs) {
        // summary: 
        //    See dojo.data.api.Read.fetch()
        //
        // description: 
        //
        //    keywordArgs.query is optional and specifies query parameters as
        //    name value pairs that are tacked on to the end of the resource uri
        //    value is a valid suffix to the 'contextRoot' supplied to the
        //    store's constructor.  For example, 'contextRoot' is set to
        //    "http://localhost:8081", then valid keywordArgs.query values are
        //    "/resources/employees", "/resources/employees/1", and
        //    "/resources/employees?firstname_contains=e" (assuming that
        //    "employees" is a defined ZRM resource. See
        //    http://www.projectzero.org/wiki/bin/view/Documentation/ResourceModel
        //
        //    If the query requires that the result set be sorted, the client
        //    must ensure that the elements of the keywordArgs.sort Array be in
        //    the desired sort order: i.e., the element specifying the 'name'
        //    attribute must precede the element specifying the 'value'
        //    attribute in order that the result set first be sorted by 'name'
        //    and then by 'value'.
        // 
        //    NOTE: The optional queryOptions parameter is irrelevant to ZRM
        //    since ZRM does not have any way to respond to its configuration
        //    directive.  We therefore ignore this parameter even if the client
        //    does supply it.
        //
        //    The client may set their own properties on the keywordArgs object
        //    (as long as they don't conflict with those reserved for ZRMStore)
        //    and they are passed-through by the ZRMStore implementation on the
        //    'request' object.
        // 

        var request = keywordArgs; 
        // see dojo.data.Request API
        request.abortedTheRequest = false;
        request.abort = function() {
            console.warn ('request.abort() fired');
            request.abortedTheRequest = true;
        };

        // first build an error handler to deal with bad client parameters
        this._errorHandler = this._buildErrorHandler(keywordArgs, request);

		var urlArgs = {_numberOfArgs:0};

        // We need server-side work to do start and count processing
        this._processStartKeyword(urlArgs, keywordArgs, request);
        this._processCountKeyword(urlArgs, keywordArgs, request);
        this._processSortKeyword(urlArgs, keywordArgs, request);
		
        var ZRMServerURL = this._buildZRMServerURL(keywordArgs, urlArgs, request);
        console.debug ('url used in fetch(): '+ZRMServerURL);
				
        this._sendQueryToZRMServer(ZRMServerURL, request);

        return (request);
    },

    _assert: function(/* boolean */ condition, /* String */ msg) {
        if (!condition) 
            throw new Error("_assert failed: "+msg);
    },

    _assertIsItem: function(/* item */ item, msg) {
        // summary:
        //
        //      Throws an exception if the item is in this ZRMStore..
        if(!this.isItem(item)) { 
            throw new Error("An zero.resource.DataStore function was passed a non-ZRMStore-item ["+msg+"]");
        }
    },
     
     _getMyItemTag: function() 
     {
       // summary:
       //   returns current value of the tag used to indicate that a given item
       //   comes from my ZRMStore instance.
       return (this._id + this._fetchCount);
     },

     isItem: function(/* anything */ candidate) 
     {
       // summary: 
       //	See dojo.data.api.Read.isItem()
       //

       // @fixme what if the item has been deleted by the client?
       if (candidate && (candidate.isZRMServerItem == this._getMyItemTag()))
	 return (true);
       else
	 return (false);
     },
     
     isItemLoaded: function(/* anything */ something)
     {
       // summary: 
       //    See dojo.data.api.Read.isItemLoaded()
       //

       // ZRMStore items are always loaded ...
       return this.isItem(something); // boolean
     },
     
     loadItem: function(/* object */ keywordArgs)
     {
       // summary: 
       //    See dojo.data.api.Read.loadItem()
       //

       // ZRMStore items are always loaded ...  "If a call to to isItemLoaded()
       // returns true before loadItem() is even called, then loadItem() need
       // not do any work at all and will not even invoke the callback
       // handlers."

       if (!keywordArgs.item)
	 throw new Error
	   ('Cannot load item: input is missing mandatory "item" property');

       this._assertIsItem(keywordArgs.item, 'loadItem()');
     },

     getValue: function(/* item */ item, 
			/* attribute-name-string */ attribute, 
			/* value? */ defaultValue)
     {
       // summary: 
       //    See dojo.data.api.Read.getValue()
       var values = this.getValues(item, attribute);
       if (values.length > 0)
	 return (values[0]);
       else
	 return (defaultValue);
     },
     
     getValues: function(/* item */ item, 
			 /* attribute-name-string */ attribute)
     {
       // summary: 
       //    See dojo.data.api.Read.getValues()
       //

       // This is also the correct return value if item does not have
       // this attribute
       var retVal = []; 
       if (this.hasAttribute(item, attribute))
	 retVal.push(item[attribute]);
       return (retVal);		// an Array
     },
     
     getAttributes: function(/* item */ item)
     {
       // summary: 
       //    See dojo.data.api.Read.getAttributes()
       this._assertIsItem(item, 'getAttributes()');
       var retVal = [];
       for (var p in item) {
	 // skip over default object properties
	 if (!minimalZRMStoreObject[p]) 
	   retVal.push(p); 
       }
       
       return (retVal);
     },

     hasAttribute: function(/* item */ item,
			    /* attribute-name-string */ attribute) 
     {
       // summary: 
       //   See dojo.data.api.Read.hasAttribute()
       //
       this._assertIsItem(item, 'hasAttribute()');

       if (!dojo.isString(attribute))
	 throw new Error
	   ("hasAttribute() supported only for attribute strings");

       if (({})[attribute] !== undefined)
	 return false;

       if (item[attribute] !== undefined)
	 return (true);
       else 
	 return (false);
     },

     containsValue: function(/* item */ item, 
			     /* attribute-name-string */ attribute, 
			     /* anything */ value)
     {
       // summary: 
       //    See dojo.data.api.Read.containsValue()
       //
       this._assertIsItem(item, 'containsValue()');

       if (!dojo.isString(attribute))
	 throw new Error
	   ("containsValue() supported only for attribute strings");

       var retVal = false;
       // an ZRMStore item can only have a single value for a given attribute
       var actualValue = this.getValue(item, attribute);
       if (actualValue) {	// item has a value for that attribute
	 // check whether that value matches the client's value
	 if (dojo.isString(value) && dojo.isString(actualValue))
	   retVal = (actualValue == value); // Strings are compared by 'value'
	 else
	   retVal = (actualValue === value); // objects, primitives
       }

       return (retVal);
     },

     _sendQueryToZRMServer:  function (ZRMServerURL, request)
     {
       // summary:
       //
       //    Sends a query to the ZRMServer, assuming that the return
       //    message will use JSON format.  We further assume that the
       //    JSON message corresponds to a legal query response.

       var getArgs = {
	 url: ZRMServerURL,
	 handleAs: "json-comment-optional",
	 sync: false		// async call
       };


       // [{"updated":"2008-01-03","id":1,"firstname":"Jerry","phonenumber":"1-XXX-XXX-XXXX","location":"North","lastname":"Cuomo"},
       // {"updated":"2008-01-03","id":100,"firstname":"Ramesh","phonenumber":"3333","location":"Downstairs","lastname":"Gopinath"}]

       // _zrmSchema.type is the cached resource type information;
       // _resourceNameInCurrentFetch is the resource type of the current query
       var needToFetchMetaData = (this._zrmSchema.type != this._resourceNameInCurrentFetch);
       var getHandler = dojo.xhrGet(getArgs);
       var zrmStore = this;

       if (needToFetchMetaData) {
	 // if fetch succeeds, follow-up with a (sucessfull) metadata query
	 getHandler.addCallback
	   (function(data) {
	     if (!zrmStore._baseZRMCollectionURI)
	       throw new Error ('_baseZRMCollectionURI is undefined');
	     var baseZRMResourceURI = 
	       zrmStore._extractBaseZRMResourceURI(zrmStore._baseZRMCollectionURI);
	     
	     var metaDataURI = baseZRMResourceURI + "/types/" + zrmStore._resourceNameInCurrentFetch;
	     //console.warn ('metadatauri=',metaDataURI);
	     zrmStore._sendMetaDataQueryToZRMServer(zrmStore, metaDataURI, data, request);
	   });
       }
       else {			// we can use cached schema information
	 getHandler.addCallback
	   (function(data) {
	     try {		// zero-out the map for a new set of items
	       zrmStore._init(zrmStore); // but NOT _initMetaData
	       zrmStore._fetchCount++;
	       zrmStore._processResultSet(zrmStore, data);
	   
	       // zrmStore._resultSet is an array of tuples: pass it to the
	       // 'clientResultSetHandler' for processing the 'items' as per the
	       // dojo client's wishes
	       var clientResultSetHandler = 
		 zrmStore._buildClientResultSetHandler(zrmStore._resultSet, request);
	       clientResultSetHandler(zrmStore._resultSet, request);
	     }
	     catch (e) {
	       zrmStore._errorHandler(e, request);
	     }
	   });
       }

       getHandler.addErrback
	 (function(error){
	   console.warn ('xhr error handler in fetch() fired: error=', error);
	   zrmStore._errorHandler(error, request);
	 });
     },

     _sendMetaDataQueryToZRMServer:  function 
       (zrmStore, metaDataURI, items, clientFetchRequest)
     {
       // summary:
       //
       //    Sends a meta-data query to the ZRMServer, assuming that the return
       //    message will use JSON format.  We further assume that the JSON
       //    message corresponds to a legal query response.  This method is
       //    invoked upon a successful completion of a prior resource query to
       //    the ZRM server, and the implementation reflects the fact that if
       //    anything goes wrong, the error has to be reflected to the error
       //    handler of the original query.

       // console.debug ('_sendMetaDataQueryToZRMServer: items=',items);

       var clientResultSetHandler = 
	 zrmStore._buildClientResultSetHandler(items, clientFetchRequest);

       var getArgs = {
	 url: metaDataURI,
	 handleAs: "json-comment-optional",
	 sync: false		// async call
       };

       //console.warn ('meta data query: url=',metaDataURI);
       // http://localhost:8080/resources/types/namevaluepairs

       /***
	   {"type":"namevaluepairs","fields":{"id":{"max_length":null,"increment":1,"start":100,"choices":[],"name":"id","primary_key":true,"type":"AutoField","required":false,"editable":true},"updated":{"max_length":null,"choices":[],"name":"updated","primary_key":false,"type":"DateTimeField","required":false,"editable":true},"name":{"max_length":50,"choices":[],"name":"name","primary_key":false,"type":"CharField","required":false,"editable":true},"value":{"max_length":50,"choices":[],"name":"value","primary_key":false,"type":"CharField","required":false,"editable":true}}}
       ***/

       var getHandler = dojo.xhrGet(getArgs);
       getHandler.addCallback
	 (function(data) {
	   
	   try {
	     // zero-out the map for a new set of items
	     zrmStore._init(zrmStore);
	     zrmStore._initMetaData(zrmStore);
	     zrmStore._fetchCount++;
	     // IMPORTANT: We need to process the meta data *before* the result
	     // set because schema info defines item identity (among onther
	     // things)
	     zrmStore._processMetaData(zrmStore, data);
	     zrmStore._processResultSet(zrmStore, items);
	     // zrmStore._resultSet is an array of tuples: pass it to the
	     // 'clientResultSetHandler' for processing the 'items' as per the
	     // dojo client's wishes
	     clientResultSetHandler(zrmStore._resultSet, clientFetchRequest);
	   }
	   catch (e) {
	     zrmStore._errorHandler(e, clientFetchRequest);
	   }

	 });
       
       getHandler.addErrback
	 (function(error) {
	   console.warn ('xhr error handler of metaData query fired: error='+error);
	   zrmStore._errorHandler(error, clientFetchRequest);
	 });

     },

     _processMetaData: function (zrmStore, metaData)
     {
       // summary:
       //
       //   processes the meta data associated with the items fetched from ZRM
       console.warn ('entering _processMetaData');
       zrmStore._assert (metaData, "metaData is undefined");
       zrmStore._assert (metaData.type, "metaData.type is undefined");
       zrmStore._assert (metaData.fields, "metaData.fields is undefined");

       zrmStore._zrmSchema.type = metaData.type;
       // not sure whether we want to process this further: may want to simply
       // add functions that use the data-structure (an object whose property
       // names are the set of item property names.  Each of these item
       // property names are, in turn, associated with schema information about
       // the specified item property.
       //
       // For example:
       // fields.{id,name,value}.{type,required,editable,primary_key,max_length}

       for (var field in metaData.fields) {
	 zrmStore._zrmSchema[field] = true;
	 // console.warn ('field', field);
	 if (metaData.fields[field].primary_key) {
	   // console.warn ('found a primary key field', field);
	   zrmStore._identityAttributes.push(field);
	 }
       }

       zrmStore._metaData = metaData.fields;
       console.warn ('exiting _processMetaData');
     },

     _buildClientResultSetHandler: function(items, clientFetchRequest)
     {
       // summary:
       //
       //    builds an handler that will be invoked if the 'fetch' operation
       //    succeeds -- so that the server returns a set of 'item's to the
       //    client's 'request'

       var scope = clientFetchRequest.scope || dojo.global;

       var clientResultSetHandler = function (items, clientFetchRequest) {
	 if (clientFetchRequest.onBegin) {
	   clientFetchRequest.onBegin.call(scope, items.length, clientFetchRequest);
	 }
	 
	 if (clientFetchRequest.onItem) {
	   for (var i = 0; i < items.length; i++) {
	     var item = items[i];
	     if (!clientFetchRequest.abortedTheRequest) // client may invoke at any time
	       clientFetchRequest.onItem.call(scope, item, clientFetchRequest);
	   }
	 }
	 
	 if (clientFetchRequest.onComplete && !clientFetchRequest.abortedTheRequest) {
	   if (clientFetchRequest.onItem) 
	     clientFetchRequest.onComplete.call(scope, null, clientFetchRequest);   
	   else
	     clientFetchRequest.onComplete.call(scope, items, clientFetchRequest);   
	 }
       };

       return (clientResultSetHandler);
     },

     _buildErrorHandler: function(keywordArgs, request)
     {
       // summary:
       //
       //   description: builds the error handler that will be invoked if
       //   "there is any sort of error while attempting to execute the query".
       var scope = request.scope || dojo.global;
       var errorHandler = null;
       if (!request.onError) {	// client didn't specify one
	 // Even though the client has (implicitly) asked us not to report any
	 // errors or do any error handling, we must throw an error because
	 // ZRMStore simply can't continue processing
	 errorHandler = function(error, request /* ignored */) {
	   console.warn ('zrmstore error handler fired: error='+error);
	   throw new Error ("Unrecoverable error occurred during fetch: "+e);
	 };
       }
       else {
	 errorHandler = function(errorData, request){
	   // "If an onError callback function is provided, the callback
	   // function will be called if there is any sort of error while
	   // attempting to execute the query."
	   request.onError.call(scope, errorData, request);
	 }
       };
       
       return (errorHandler);
     },

     getFeatures: function()
     {
       // summary: 
       //    See dojo.data.api.Read.getFeatures()
       return (this._features);
     },

     close: function(/*dojo.data.api.Request || keywordArgs || null */ request)
     {
       // summary: 
       //     See dojo.data.api.Read.close()
       //
       // description: 
       //   This method is a no-op for ZRMStore because the next query will
       //   flush the existing set of items automatically, and the current set
       //   of items are inaccessible except through the fetch callbacks.  No
       //   other state is maintained by the store: e.g., ZRMStore does not
       //   maintain any 'open' connections.
     },

     getLabel: function(/* item */ item)
     {
       //	summary: 
       //		See dojo.data.api.Read.getLabel()

       // "or undefined if no user-readable label can be generated" ...
       //
       // A ZRM Server has no knowledge about what attributes might be used as
       // a label.  We may revisit this decision when we integrate use of
       // "primary keys" into the implementation (e.g., Identity")

       return (undefined); 
     },
     
     getLabelAttributes: function(/* item */ item)
     {
       //	summary: 
       //		See dojo.data.api.Read.getLabelAttributes()

       // "or null if public attributes were not used to generate the label."
       //
       // A ZRM Server has no knowledge about what attributes might be used as
       // a label.  We may revisit this decision when we integrate use of
       // "primary keys" into the implementation (e.g., Identity")

       return (null);
     },
     
     _processCountKeyword: function(urlArgs, keywordArgs, request)
     {
       // summary:
       //
       //    Inserts the optional 'count' keyword property appropriately into
       //    the set of urlArgs
       var countKeyword = keywordArgs.count;

       if (countKeyword !== undefined) { // optional keyword
		 // not exactly what I want since "3" is just as valid as 3
		 var countValue = parseInt(countKeyword);

		if (isNaN(countValue))
	   		this._errorHandler
	     		(new Error("Invalid 'keywordArgs.count' value: "+countKeyword),
	      		request);

	 	if (countValue < 0)
	   	this._errorHandler
	     	(new Error("Value of 'keywordArgs.count' may not be negative :"+
				countValue), request);

	 	urlArgs[this.ZRMConstants.countParam] = countValue;
	 	this._incrementNumberOfURLArgs(urlArgs);
       }
     },

     _processStartKeyword: function(urlArgs, keywordArgs, request)
     {
       // summary:
       //
       //    Inserts the optional 'start' keyword property appropriately into
       //    the set of urlArgs
       var startKeyword = keywordArgs.start;
       if (startKeyword !== undefined) { // optional keyword
		 // not exactly what I want since "3" is just as valid as 3
		 var startValue = parseInt(startKeyword);

	 	if (isNaN(startValue))
	   		this._errorHandler
	     		(new Error("Invalid 'keywordArgs.start' value: "+startKeyword),
	      		request);

	 	if (startValue < 0)
	   		this._errorHandler
	     		(new Error
	      			("Value of 'keywordArgs.start' may not be negative :"+
	       			startValue), request);

	 	urlArgs[this.ZRMConstants.startParam] = startValue;
	 	this._incrementNumberOfURLArgs(urlArgs);
       }
     },

     _processSortKeyword: function(urlArgs, keywordArgs, request)
     {
       // summary:
       //
       //    Inserts the optional 'sort' keyword property appropriately into
       //    the set of urlArgs
       var sortKeyword = keywordArgs.sort;
       if (sortKeyword) {	// optional keywordArgs property
	 	if ((typeof sortKeyword != "array") &&
	     (!(sortKeyword instanceof Array)))
	   		this._errorHandler
	     		(new Error("Invalid 'keywordArgs.sort' value: must be an array, "+
					"is a "+(typeof sortKeyword)), request);

	 	if (sortKeyword.length < 1)
	   		this._errorHandler
	     		(new Error("Invalid 'keywordArgs.sort' value: "+
					"must have at least one element"));

	 	var sortDirective = "";
	 	for (var i=0; i<sortKeyword.length; i++) {
	   		var sortElement = sortKeyword[i];
	   		var attribute = sortElement.attribute;
	   		if (typeof attribute != "string")
	     		this._errorHandler
	       			(new Error ("The keywordArgs.sort property's 'attributes' must"+
			   		"be string-valued"), request);
	   		var descending = false;
	   		if (sortElement.descending !== undefined) // watch those booleans
	     		descending = sortElement.descending;
	   		if (typeof descending != "boolean")
	     		this._errorHandler
	       		(new Error ("The keywordArgs.sort property's 'descending'" +
			   		"value must be boolean valued"), request);
	   
	  	 	if (descending)
	     		sortDirective += "-"; // ascending is default for ZRM 
	   		sortDirective += attribute;
	   		if (i < sortKeyword.length -1) // at least one sort element left
	     		sortDirective += ",";
	 	}

	 	urlArgs[this.ZRMConstants.sortAttribute] = sortDirective;
	 	this._incrementNumberOfURLArgs(urlArgs);
	 
       } // if sortKeyword
     },

     _processResultSet:  function(theStore, data)
     {
       // summary:
       //
       //   An internal function that processes the result set that's sent by
       //   the ZRM Server.
       //
       //   (1) Builds a data structure that maps each item's
       //   identity to the item itself.
       //
       //   (2) Tags each item as being a "ZRMServer" item of this ZRMStore
       //   instance (not a plain array): see "isItem()" etc in the
       //   dojo.data.read API for our motivation
       //
       //   (3) Clears each item's "dirty" bit 
       //
       // NOTE: We assume that identityToItem is an empty array when it enters
       // this function

       theStore._resultSet = data;

       var myItemTag = theStore._getMyItemTag();

       for (var i=0; i < data.length; i++) {
	 var item = data[i];

	 // First, tag the item as ours
	 // @fixme: be nice if we could reuse '_enlistAnItem'
	 item.isZRMServerItem = myItemTag; 

	 item.dirtyBit = false;

	 var identity = theStore.getIdentity(item);
	 theStore._identityToItem[identity] = item;

       }
     },

    _buildZRMServerURL:  function(keyWordArgs, urlArgs, request) {
        // summary:
        //
        //    Returns a URL in accordance with the REST API required by
        //    the ZRMServer
		var queryParams = {};
		if (keyWordArgs.query) {
			if (typeof keyWordArgs.query != "object") {
	   			this._errorHandler
	     			(new Error("Invalid 'keywordArgs.query' value: must be an object, "+
						"is a "+ (typeof keywordArgs.query)), request);
			} else {
				queryParams = keyWordArgs.query;
			}
		}
        var ec = encodeURIComponent;
        // we need to save it for the "save()" operation
        this._baseZRMCollectionURI = this.contextRoot + "/resources/" + this.resourceCollection ;

        // cache to determine whether we can use cached schema information
        this._resourceNameInCurrentFetch =  this._extractResourceFromBaseZRMCollectionURI(this._baseZRMCollectionURI);

        // the actual url may well include filter conditions
        var theURL = this._baseZRMCollectionURI;

		// Build up the query string
		var queryString = "";
        for (var x in queryParams) {
            if (x == "_numberOfArgs" || queryParams[x] == "*") 
                continue;
            queryString += ec(x) + '=' + ec(queryParams[x]) + '&';
        }
        
        for (var x in urlArgs) {
            if (x == "_numberOfArgs") 
                continue;
            queryString += ec(x) + '=' + ec(urlArgs[x]) + '&';
        }
       
        // get rid of the last ampersand
        if (queryString.charAt(queryString.length-1) == "&")
            queryString = queryString.substr(0, queryString.length-1);

		// Check if the resourceCollection included a query            
        var URLDoesntContainQuestionMark =  (-1 === theURL.indexOf('?'));
        if (queryString.length > 0) { // at least one filter, sort directive etc
	       if (URLDoesntContainQuestionMark) {
	           theURL += '?' + queryString;
	       } else {
	           // url contains a question mark: e.g.,
	           // http://localhost:8080/resources/namevaluepairs?name__equals=name1
	           theURL += '&' + queryString; // to deal with subsequent stuff like start=0&count=1
	       }
        }
        
        return (theURL);
    },
    
    _incrementNumberOfURLArgs: function (urlArgs)
    {
		urlArgs._numberOfArgs++;
    },

    _getNumberOfURLArgs: function (urlArgs)
    {
		return (urlArgs._numberOfArgs);
    },
		
     // ----------------------------------------------------------------------------
     // dojo.data.Identity	      
     // ----------------------------------------------------------------------------
     getIdentity: function (/* item */ item)
     {
       // summary: 
       //    See dojo.data.api.Identity.getIdentity()
       //
       // description: We return a string that is the concatenation of the
       // value of the set of identity attributes.  We've been told that ZRM
       // identity is specified by exactly one property: our implementation
       // works for the case where 'getIdentityAttributes()' returns an array.
       // We've also been told that ZRM property values are not objects, and
       // our implementation asserts this to be true.

       this._assertIsItem(item, 'getIdentity()');
       
       var identity = "";

       if (item.pseudoIdentity)	// it's one of our items, but created by us
	 throw new Error
	   ("getIdentity() can only be invoked on an item fetched from the "+
	    "server: your item was created by the ZRMStore");
       else {
	 var identityAttrs = this.getIdentityAttributes(item);
       
	 for (var i=0; i<identityAttrs.length; i++) {
	   var identityAttribute = identityAttrs[i];
	   var identityAttributeValue = 
	     this.getValue(item, identityAttribute);
	   this._assert
	     ((typeof identityAttributeValue != "object"), 
	      "identity property "+identityAttribute+" is not an object");
	   identity += identityAttributeValue; // should work for int, boolean etc
	   identity += "_";
	 }
       }       

       return (identity);
     },
	      
     _getInternalIdentity: function (/* item */ item)
     {
       // summary: internal function that returns an item's identity -- as
       // defined and used by the ZRMStore's internal data-structures.  Unlike
       // "getIdentity()" which only returns the identity of items fetched from
       // the server, this function also returns the identity of items created
       // and not yet saved by the ZRMStore.

       this._assertIsItem(item, '_getInternalIdentity()');

       var identity = null;

       if (item.pseudoIdentity)
	 identity = item.pseudoIdentity;
       else
	 identity = this.getIdentity(item);

       return (identity);
     },

     getIdentityAttributes : function (/* item */ item)
     {
       // summary:
       //   See dojo.data.api.Identity.getIdentityAttributes()
       //
       // description: we use the ZRM meta-data to access the primary_key
       // information (aka identity attributes), and cache it for future
       // access.  This means that as long as the item is a valid item, we
       // don't need to access the item to get the necessary information.

        this._assertIsItem(item, 'getIdentityAttributes()');
	return (this._identityAttributes);
     },

     fetchItemByIdentity: function(/* object */ keywordArgs) {
        // summary:
        //   See dojo.data.api.Identity.fetchItemByIdentity
        //
        // description: this implementation looks only in the local cache for
        // the item and does NOT query the ZRM server on a cache-miss.  We may
        // want to change this behavior in later releases.
        //
        // IMPORTANT: because identity is defined only by the ZRM server, it is
        // impossible for the client to fetch a newly-created item (one created
        // on the client) by its identity.  Only items fetched from the server
        // can be 'fetched by identity'.

        // TODO need to figure out how to get this from metadata
        // keywordArgs.identity = "id";
         
        var item = null;

        if (!keywordArgs.identity)
            keywordArgs.identity = "id";
            // throw new Error("'keywordArgs.identity' is a mandatory parameter for "+"'fetchItemByIdentity'");

        // null if not there
        item = this._lookupItemInLocalCache(keywordArgs.identity, "fetchItemByIdentity()");

        // No processing of "keywordArgs.onError" parameter because it "is the
        // callback to invoke when the item load encountered an error."  Our
        // implementation looks only on the local cache so there can be no error
        // loading the item.
        if (keywordArgs.onItem) {
            var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
            keywordArgs.onItem.call(scope, item);
        }
    },

    _lookupItemInLocalCache : function (itemIdentifier, callerContext) {
        // summary:
        //
        // Internal function to look up an item by its "identity" which must be
        // of type String.
        //
        // Returns an item or null as appropriate
        //
        // IMPORTANT: because identity is defined only by the ZRM server, it is
        // impossible for the client to fetch a newly-created item (one created
        // on the client) by its identity.  Only items fetched from the server
        // can be 'fetched by identity'.

        this._assert(dojo.isString(itemIdentifier), 'item identifier '+itemIdentifier+' is not a string');
        var theItem = this._identityToItem[itemIdentifier];
        if (theItem)
            return (theItem);
        else
            return (null);
    },
 
     // ----------------------------------------------------------------------------
     // dojo.data.Write	      
     // ----------------------------------------------------------------------------
     setValue: function(/* item */ item, 
 			/* string */ attribute,
  			/* almost anything */ value)
     {
       // summary: 
       //    See dojo.data.api.Write.setValue()
       //
       // description: assuming the operation is valid, we save a copy of the
       // original item in the 'identityToOriginal" map, mark the item in the
       // 'resultSet' as dirty, and then modify the 'resultSet' item. 
       //
       // IMPORTANT: The spec allows this method to be used to *add* new
       // attributes.  We forbid this, because the ZRM server associates
       // resources with fixed schema
  
       /**
       this._assert (!this._saveInProgress, 
		     "Save operation currently in progress");
       */

       this.setValues(item, attribute, [value]);
     },
 
     _setValue: function(/* item */ item, 
			 /* string */ attribute,
			 /* almost anything */ value)
     {
       // summary: Internal function to save the specified value for the item's
       // attribute: differs from "setBalue()" in that all parameters are
       // assumed to be validated.

       if (item.pseudoIdentity)	
	 // Since client didn't do a save yet, the set of new items holds the
	 // modified state rather than the original state of the item.  We
	 // don't need to do a backup of the old value of the new item
	 ;
       else {
	 // make a backup copy of the current state of the item
	 var copiedItem = {};	 

	 // IMPORTANT: This is not a deep-copy -- i.e., if item has object,
	 // array, or function valued properties, they will be copied only by
	 // 'reference'.  We have been assured by the ZRM team that item
	 // properties will not be arrays or objects.
	 for (var p in item)
	   copiedItem[p] = item[p];

	 this._identityToOriginalItem[this.getIdentity(item)] = copiedItem;
	 // add the modified item to the set of modified items
	 this._identityToModifiedItem[this.getIdentity(item)] = item;
       }

       // change the state of the original item and make it available as part
       // of the working set: the map of modified items will store this changed
       // state well.
       var oldValue = item[attribute];
       item[attribute] = value;
       item.dirtyBit = true;
       
       // implement the dojo.data.api.Notification call
       this.onSet(item, attribute, oldValue, value);
     },
 
     setValues: function(/* item */ item,
			 /* string */ attribute, 
			 /* array */ values)
     {
       // summary: 
       //    See dojo.data.api.Write.setValues()

       /**
       this._assert (!this._saveInProgress, 
		     "Save operation currently in progress");
       */

       this._assertIsItem(item, 'setValues()');
       this._assert(dojo.isArray(values),
		    "'values' parameter to setValues must be an array");
       this._assert(this.hasAttribute(item, attribute),
		    "attribute "+attribute+" not part of ZRM item's schema");
       this._assert(!(this._isIdentityAttribute(attribute)),
		    "Illegal to set the value of an identity attribute: "+attribute);

       if (values.length === 0) {
	 // "Calling store.setValues(x, y, []) (with *values* as an empty
	 // array) has the same effect as calling store.unsetAttribute(x, y)."
	 this.unsetAttribute(item, attribute);
       }
       else if (values.length === 1) {
	 var value = values[0];
	 this._assert(value !== undefined,
		      'value parameter is undefined');
	 this._setValue(item, attribute, value);
       }
       else {
	 throw new Error
	   ('ZRMStore does not support setting attributes to non-scalar '+
	    'values: you supplied '+values.length+' values for attribute '+
	    attribute);
       }
     },

     unsetAttribute: function(/* item */ item, /* string */ attribute)
     {
       // summary: 
       //    See dojo.data.api.Write.unsetAttribute()
       //
       // description: 
       //
       //   our implementation sets the attribute's value to "null" rather than
       //   "unset" or "delete" the property from the object (since ZRM
       //   requires items to conform to a fixed schema).
       //   
       // NOTE: If the client sets a property to null via "setValue()", we and
       // ZRM are unable to distinguish that from "unsetting" the attribute.
       //
       // Some of our reasoning behind this approach: Property "foo" not
       // existing is different from property "foo" having a null value in
       // Javascript, and this seems to carry over to JSON -- {a: 42} vs {a:
       // 42, foo: 43} However, a relational database does not make this
       // distinction (you can't make columns go away for individual rows), so
       // it's necessary to (a) ignore the distinction, or b) kludge something
       // on the database side, which is ugly and impractical.  This
       // implementation follows the approach (a).
       //
       // NOTE: 'undefined' is a legal value in JavaScript.  Although the Dojo
       // toJson() function does seem to propagate 'undefined' to JSON,
       // json.org (http://www.json.org/js.html) says that undefined properties
       // are not supposed to show up in JSON.  This implementation takes the
       // approach of assuming that the 'undefined' value will not appear in
       // JSON.

       this._assertIsItem(item, 'unsetAttribute()');
       this._assert(this.hasAttribute(item, attribute),
		    "attribute "+attribute+" not part of ZRM item's schema");
       this._assert(dojo.isString(attribute), 
		    'attribute '+attribute+' is not a string');

       item[attribute] = null;
     },

     revert: function() 
     {
       // summary:
       //    See dojo.data.api.Write.revert()
       // 

       // @fixme if we simply ignore the _resultSet ds once the fetch has
       // completed, and just maintain a set of "identity to current, original,
       // new, and deleted" mappings -- then we can implement 'revert' by
       // simply getting clearing the map of modified items

       // We can't control the set of object references that the client may
       // have: all we can do is restore the state of the ZRMStore for
       // subsequent client access to the store (e.g.,
       // "fetchItemByIdentity()").  So, we iterate over the set of modified
       // items, and restore the state of the items that we access when the
       // client accesses the store to their original values.
       for (var itemIdentity in this._identityToOriginalItem) {
	 var originalItem = this._identityToOriginalItem[itemIdentity];
	 // cache lookups use the '_identityToItem' map
	 var modifiedItem = this._identityToItem[itemIdentity];

	 // revert the item's properties to their original values
	 for (var p in originalItem) // note shallow-copy notes in "setValue"
	   modifiedItem[p] = originalItem[p];

	 modifiedItem.dirtyBit = false;
       }

       // Before clearing the set of items created on the client, we have to
       // change the state of formerly new items (now toast) to reflect the
       // fact that they're not legitimate items anymore.
       for (var itemIdentity in this._identityToNewItem)
	 this._delistAnItem(this._identityToNewItem[itemIdentity]);
       this._identityToNewItem = {};

       // clear the set of copies of the original items since the original
       // state has been restored
       this._identityToOriginalItem = {};

       // @fixme need to deal with reverting items that were previously deleted
       // by the client.  Need to restore from _identityToDeletedItem map, but
       // somehow have to track whether original item was created on client
       // versus fetched from server.  Also need to remove item from
       // _identityToDeletedItem map

       for (var identity in this._identityToDeletedItem) {
	 // we don't have to revert deleted items that were created on the
	 // client and not saved to the server
	 var currentlyDeletedItem = this._identityToDeletedItem[identity];
	 this._enlistAnItem(currentlyDeletedItem);
	 this._identityToItem[identity] = currentlyDeletedItem;
       }
       // all deleted items have been 'un-deleted'
       this._identityToDeletedItem = {};

       // no more modified items
       this._identityToModifiedItem = {};

       return (true);
     },

     deleteItem: function(/* item */ item)
     {
       // summary:
       //    See dojo.data.api.Write.deleteItem()
       // 

       console.warn("ZRMS.deleteItem", item);

       this._assertIsItem(item, 'deleteItem()');
       /*
       this._assert (!this._saveInProgress, 
		     "Save operation currently in progress");
       */

       // @fixme do we need to delete the item from the _resultSet?

       // may be an item created by us
       var identity = this._getInternalIdentity(item);

       // add to the set of deleted items -- but only if it wasn't an newly
       // created item: such items just get rolled back
       if (item.pseudoIdentity)
	 delete (item.pseudoIdentity);
       else {
	 // we set the state of the deleted item to the *original* state of the
	 // item so that 'revert()' can work if subsequentally called
	 var originalItem = this._identityToItem[identity];
	 this._identityToDeletedItem[identity] = originalItem;
       }

       // let's not get complicated: just delete the item from any
       // data-structure in which it may lurk ('delete' is a no-op if the
       // object property doesn't exist)
       delete (this._identityToNewItem[identity]);
       delete (this._identityToItem[identity]);
       delete (this._identityToOriginalItem[identity]);
       delete (this._identityToModifiedItem[identity]);

       this._delistAnItem(item);

       // implement the dojo.data.api.Notification call
       this.onDelete(item);

       return (true);
     },

     save: function(/* object */ keywordArgs)
     {
       // @fixme need to fence this and other functions so that clients can
       // add/modify/delete items while save is in flight

       this._assert (keywordArgs,
		     "save() must be passed a non-null/undefined keywordArgs");

       /**
       this._assert (!this._saveInProgress, 
		     "Save operation currently in progress");
       this._saveInProgress = true;
       */

       var numberOfUncompletedSavedItems = {'howManyLeft': 0};
       var anySaved = false;

       // @fixme we're about to fire off a whole mess of asynchronous calls:
       // will this cause problems, do we need to throttle them etc?

       for (var identity in this._identityToDeletedItem) {
	 numberOfUncompletedSavedItems.howManyLeft++;
	 anySaved = true;
	 this._saveItem("xhrDelete", this._identityToDeletedItem[identity],
			keywordArgs, numberOfUncompletedSavedItems);

	 if (numberOfUncompletedSavedItems.errorFired) {
	   console.warn 
	     ('In save(): stopped processing of deleted items '+
	      'because an error occured');
	   break;
	 }
       }

       for (var identity in this._identityToModifiedItem) {
	 numberOfUncompletedSavedItems.howManyLeft++;
	 anySaved = true;
	 this._saveItem("rawXhrPut", this._identityToModifiedItem[identity],
			keywordArgs, numberOfUncompletedSavedItems);

	 if (numberOfUncompletedSavedItems.errorFired) {
	   console.warn 
	     ('In save(): stopped processing of modified items '+
	      'because an error occured');
	   break;
	 }
       }

       for (var identity in this._identityToNewItem) {
	 numberOfUncompletedSavedItems.howManyLeft++;
	 anySaved = true;
	 this._saveItem("rawXhrPost", this._identityToNewItem[identity],
			keywordArgs, numberOfUncompletedSavedItems);

	 if (numberOfUncompletedSavedItems.errorFired) {
	   console.warn 
	     ('In save(): stopped processing of new items '+
	      'because an error occured');
	   break;
	 }
       }

       if (!anySaved)	{	// no new, modified, or deleted item
	 if (keywordArgs.onComplete) {
	   var scope = (keywordArgs.scope || dojo.global);
	   // synchronous call invokes handler since the save has "completed"
	   keywordArgs.onComplete.call(scope);
	 }
       }
       
     	 
       /**
	  this._saveInProgress = false;
       */
     },

     isDirty: function (/* item? */ item)
     {
       // summary: 
       //    See dojo.data.api.Write.isDirty()

       // description: our implementation considers an item to be
       // "modified" iff setValue(s) has been invoked on it.  the
       // store is considered "modified" iff one of its items is
       // dirty.

       if (item) {
	 this._assertIsItem(item, 'isDirty()');
	 return (item.dirtyBit);
       }
       else {
	 // "If isDirty() is called with no *item* argument, then this function
	 // returns true if any item has been modified since the last save()."
	 var isModifiedItem = false;
	 for (var identity in this._identityToModifiedItem) {
	   isModifiedItem = true;
	   break;
	 }
	 return (isModifiedItem);
       }
     },

    _buildXHRSaveArgs: function(operation, zrmResourceURL, data) {
        var xhrArgs = {url: zrmResourceURL, handleAs: "json-comment-optional", sync: false};
       
        if (operation == "xhrPost" || operation == "rawXhrPost" || operation == "xhrPut" || operation == "rawXhrPut") {
            if (!data)
                throw new Error(operation + " with no data");
            xhrArgs.contentType = "application/json";
            if (operation == "rawXhrPost") {
                xhrArgs.postData = dojo.toJson(data);
            } else if (operation == "xhrPost") {
                xhrArgs.content = dojo.toJson(data);
            } else if (operation == "xhrPut") {
                xhrArgs.content = dojo.toJson(data);
            } else if (operation == "rawXhrPut") {
                xhrArgs.putData = dojo.toJson(data);
            }
        }
        
        if (operation == "xhrPut" || operation == "rawXhrPut" || operation == "xhrDelete") {
            xhrArgs["handleAs"] = "text";
        }
       
       return (xhrArgs);
     },

     _saveItem: function(operation, item, keywordArgs, numberOfUncompletedSavedItems) 
     {
       var zrmResourceURI = this._baseZRMCollectionURI;
       if (operation == "xhrPost" || operation == "rawXhrPost")
	 // new item -- creating a resource: no need to identify the element in
	 // the resource collection, so just use the resource collection URI
	 ;			
       else {
	 // add trailing "/" if necessary
	 if (zrmResourceURI.charAt(zrmResourceURI.length) != "/")
	   zrmResourceURI += "/";

	 // IMPORTANT: We need to get the identity attributes, but in the case
	 // of a deleted item, cannot call 'getIdentityAttributes()' since the
	 // item is no longer an item (if you see what I mean).  So, we cheat
	 // and use the internal representation of the item to bypass the API
	 // call.
	 var identityAttrs = this._identityAttributes;
       
	 for (var i=0; i<identityAttrs.length; i++) {
	   var identityAttribute = identityAttrs[i];
	   // NOTE: the code would need fixing if they ever support multiple id
	   // attributes
	   var identityAttributeValue = 
	     // IMPORTANT: We need to get the value of the identity property
	     // but in the case of a deleted item, cannot call 'getValue()'
	     // since the item is no longer an item (if you see what I mean).
	     // So, we cheat and use the internal representation of the item to
	     // bypass the API call.
	     item[identityAttribute]; // this.getValue(item, identityAttribute);
	   zrmResourceURI += identityAttributeValue;
	 }
       }

       var data = {};
       for (var p in item) {
	 // skip over properties that we've added to ZRM
	 if (!minimalZRMStoreObject[p]) 
	   data[p] = item[p];
       }
       
       //console.warn ('_saveItem -- item=',item,', uri='+zrmResourceURI,', sent data=', data);

       var saveArgs = this._buildXHRSaveArgs(operation, zrmResourceURI, data);

       if (!dojo[operation])
	 throw new Error("Bad save() operation: " + operation);
       var saveHandler = dojo[operation](saveArgs); // e.g., dojo.xhrGet

       var scope = (keywordArgs.scope || dojo.global);
       var self = this;
       saveHandler.addCallback(function(data) {
	 numberOfUncompletedSavedItems.howManyLeft--;
	 // console.debug ('number of uncompleted items to save = ',
	 // numberOfUncompletedSavedItems.howManyLeft);

	 if (0 === numberOfUncompletedSavedItems.howManyLeft) {
	   //onComplete should only fire when the entire save is done.  Not for
	   //each individual item.
	   if (keywordArgs.onComplete) {
	     keywordArgs.onComplete.call(scope);
	 }

	   console.debug
	     ('In save: no more items to process: initiating post-save cleanup');
	   self._postSaveProcessing(self);
	 }

       });
       
       saveHandler.addErrback(function(errorData) {
	 if (keywordArgs.onError) {
	   keywordArgs.onError.call(scope, errorData);
	 }
	 
	 numberOfUncompletedSavedItems.errorFired = true;
	 console.warn ('save error handler fired', errorData);
       });
     },

     _postSaveProcessing: function(zrmStore)
     {
       // we don't zero out the "_identityToItem" map (and the "_resultSet" array)
       // because they holds the current result set.
       zrmStore._identityToOriginalItem = {};
       zrmStore._identityToNewItem = {};
       zrmStore._identityToDeletedItem = {};
       zrmStore._identityToModifiedItem = {};

       // NOTE: We clear the '_saveInProgress' flag in the 'save()' function
       // itself
     },

     newItem: function (/* Object? */ keywordArgs, /*Object?*/ parentInfo)
     {
       // summary: 
       //   See dojo.data.api.Write.newItem()
       //
       // description: This implementation forbids the client for passing an
       // item with properties that are not already part of the ZRM Server's
       // schema for the resource that's cached in this ZRMStore instance.

       /**
       this._assert (!this._saveInProgress, 
		     "Save operation currently in progress");
       */

       this._assert 
	 ((null != this._baseZRMCollectionURI && "" != this._baseZRMCollectionURI),
	  "Cannot invoke 'newItem()' before doing a fetch() so that the store "+
	  "is supplied with a URI to a ZRM Resource Collection");

       if (!dojo.isObject(keywordArgs) && keywordArgs !== undefined)
	 throw new Error
	   ("keywordArgs passed to newItem() was passed something other than "+
	    "an object");
       
       if (parentInfo)
	 throw new Error 
	   ("ZRMStore is not a hierarchical store: 'parentInfo' is therefore "+
	    " an invalid parameter to newItem()");

       var newItem = {};
       for (var attribute in keywordArgs) {

	 this._assert
	   ((this._zrmSchema[attribute] !== undefined),
	    "The ZRM schema does not include attribute "+attribute+" of new item");
	   
	 this._assert
	   (!(this._isIdentityAttribute(attribute)),
	    "Illegal to set the value of an identity attribute: "+attribute);

	 // do a shallow copy of the keyword properties: see the comments in
	 // "setValue()" about our assumptions
	 var theValue = keywordArgs[attribute];
	 this._assert
	   (theValue !== undefined, 
	    'the value of property '+attribute+' is undefined');
	 this._assert
	   (this._isScalar(theValue),
	    'ZRMStore requires that value of property '+attribute+
	    'be a scalar: you supplied an '+typeof(theValue));

	 newItem[attribute] = theValue;
       }
	
       // IMPORTANT: Only the ZRMServer can assign an identity to this item
       // (and this will happen as a side-effect of the "save()" operation if
       // the client subsequentally invokes it).  We need to assign at least a
       // "pseudo-identity" so that we can track the new item internally, but
       // throw an error if the client tries to use that identity
       newItem.pseudoIdentity = this._getAndIncrementPseudoIdCounter();
       this._enlistAnItem(newItem);

       // @fixme -- not clear whether we need to insert the new item into into
       // the resultSet data-structure since the client can't get access to the
       // ds after the 'fetch()' and a new 'fetch()' wipes the current contents
       // anyway.
	 
       // insert the item into the the identity map data structures 
       this._identityToNewItem[newItem.pseudoIdentity] = newItem;

       // implement the dojo.data.api.Notification call
       this.onNew(newItem);

       return (newItem);
     },

     _delistAnItem: function(item)
     {
       // summary: internal function that transforms an item to reflect the
       // fact that it's no longer considered a legitimate item
       this._assertIsItem(item, 'delistAnItem()');
       delete (item.isZRMServerItem);
     },

     _enlistAnItem: function(item)
     {
       // summary: internal function that transforms an item to reflect the
       // fact that it's been enlisted as a legitimate ZRMStore item

       // @fixme should we validate that not currently an item
       if (item)
	 item.isZRMServerItem = this._getMyItemTag();
     },

     // ----------------------------------------------------------------------------
     // dojo.data.api.Notification
     // ----------------------------------------------------------------------------

     onSet: function(/* item */ item, 
		     /*attribute-name-string*/ attribute, 
		     /*object | array*/ oldValue,
		     /*object | array*/ newValue)
     {
       // summary: See dojo.data.api.Notification.onSet()
		
       // No need to do anything. This method is here just so that client code
       // can connect observers to it.
     },

     onNew: function(/* item */ newItem, /*object?*/ parentInfo)
     {
       // summary: See dojo.data.api.Notification.onNew()
       
       // No need to do anything. This method is here just so that client code
       // can connect observers to it.
     },

     onDelete: function(/* item */ deletedItem)
     {
       // summary: See dojo.data.api.Notification.onDelete()
       
       // No need to do anything. This method is here just so that client code
       // can connect observers to it.
     },

     _isScalar: function (value)	       
     {
       // summary: internal function that returns true iff 'value' is neither
       // object, nor array, nor function, false otherwise.  We assume that
       // 'value' is not undefined.
       //
       // IMPORTANT: returns true if value is a string or if value is null
       
       var isScalar = false;
       if ((value === null) || (dojo.isString(value)))
	 isScalar = true;
       else if (dojo.isObject(value))
	 // dojo.isObject(): "Returns true if it is a JavaScript object (or an
	 // Array, a Function or null)"
	 isScalar = false;
       else 
	 isScalar = true;

       return (isScalar);
     },


     _getAndIncrementPseudoIdCounter : function()
     {
       return (String(this._pseudoIdCounter++));
     },

     _isIdentityAttribute: function (/* string */ attribute)
     {
       // summary: internal function that uses the ZRM meta-data to return true
       // iff 'attribute' is an identity attribute, false otherwise.
       //
       // NOTE: Unlike the dojo.data API we leverage the meta-data info to
       // obviate the need for an 'item' parameter.

       var isIdentityAttribute = false;
       for (var i=0; i<this._identityAttributes.length; i++) {
 	 var identityAttribute = this._identityAttributes[i];
	 if (attribute == identityAttribute) {
	   isIdentityAttribute = true;
	   break;
	 }
       }

       return (isIdentityAttribute);
     },

     _init: function(zrmStore)
     {
       // summary: initialize the data structures properties on 'obj' (used
       // instead of 'this' because of event-handler scoping issues)

       zrmStore._resultSet = [];
       zrmStore._identityToItem = {};
       zrmStore._identityToOriginalItem = {};
       zrmStore._identityToNewItem = {};
       zrmStore._identityToDeletedItem = {};
       zrmStore._identityToModifiedItem = {};
     },

      _initMetaData: function(zrmStore)
      {
       // schema stuff
       zrmStore._metaData = {};
       zrmStore._identityAttributes = [];
       zrmStore._zrmSchema = {};
     },

    _extractBaseCollectionURI: function(zrmCollectionURI) {
        var retVal = "";
        
        if (zrmCollectionURI)
            zrmCollectionURI = this._extractQueryURI(zrmCollectionURI);
            
        if (zrmCollectionURI) {
            var firstQuestionMark = zrmCollectionURI.indexOf("?");
            if (-1 == firstQuestionMark)
                retVal = zrmCollectionURI;
            else
                retVal = zrmCollectionURI.substring(0, firstQuestionMark);

            if (retVal.charAt(retVal.length) == "/")
                throw new Error("Error: collection URL may not end with /: " + retVal);
        }

        return (retVal);
    },

     _extractResourceFromBaseZRMCollectionURI: function(baseZRMCollectionURI)
     {
       // summary: 
       //
       //   extract the resource (e.g., "namevaluepairs") from the base ZRM
       //   Collection URI (e.g.,
       //   "http://localhost:8080/resources/namevaluepairs")

       // find the last "/" in the URI
       var lastBackslash = baseZRMCollectionURI.lastIndexOf("/");
       // extract everything past that last "/"
       return (baseZRMCollectionURI.substring
	       (lastBackslash+1, baseZRMCollectionURI.length));
     },

     _extractBaseZRMResourceURI: function(baseZRMCollectionURI)
     {
       // summary: 
       //
       //   extract the base ZRM resource URI (e.g.,
       //   "http://localhost:8080/resources") from the base ZRM Collection URI
       //   (e.g., "http://localhost:8080/resources/namevaluepairs")

       // find the last "/" in the URI
       var lastBackslash = baseZRMCollectionURI.lastIndexOf("/");
       // extract everything up to that last "/"
       return (baseZRMCollectionURI.substring(0, lastBackslash));
     },

    getMetadata: function(){
        var md = undefined;
		var schemaUri = this.contextRoot + "/resources/types/" + this.resourceCollection;
        if(!md){
            dojo.xhrGet({
                url: schemaUri,
                handleAs: "json",
                sync: true
            }).addCallback(function(result){
                md = result;
            });
        }
        this.metadata = md;
        console.debug("Schema",dojo.toJson(this.metadata));
        return this.metadata;
    },
    _defaultValue: function(md){
        //return appropriate default value for a field, based on metadata
        //FIXME: ZRM needs a better way to get an realistic default value
        return "";
    },
    createDefaultItem: function(){
        var obj = {};
        for (var fname in this.metadata.fields){
            if (fname){
                if ((this.metadata.fields[fname].type === "auto") ||
                    (this.metadata.fields[fname].readonly)){
                    continue;
                }
                obj[fname]=(this._defaultValue(this.metadata.fields[fname]));
            }
        }
        return obj;
    }
     
 });				// end the "declare"

	     
