dojo.provide("zero.resource.RestStore");

dojo.require("dojo.data.util.filter");
dojo.require("dojo.data.util.sorter");

dojo.declare("zero.resource.RestStore", null, {
	//	summary:
	//		Shim store to talk to a remote datastore.
    //
    //  based on a Store by: Jared Jurkiewicz and Ben Schell

	resourceUri: "",
	_storeRef: "_S",
	memberLabel: "", // text representation of the Item [label and identifier may need to stay due to method names]
	memberIdentifier: "", // unique identifier of the Item
	_changedItems: null, // Items that have been modified
	// The attribute on the returned JSON object to look for an array of values.  For example, 'users' or 'builds'.

	constructor: function(/* Object */ keywordParameters){
		if(keywordParameters){
			dojo.mixin(this, keywordParameters);
		}
	},

	getFeatures: function(){
		return { 
			"dojo.data.api.Read" : true, 
			"dojo.data.api.Identity" : true,
            "dojo.data.api.Notification": true,			
			"dojo.data.api.Write": true			
		};
	},

	isItem: function(item){
		if (item && item[this._storeRef] == this) {
			return true;
		}
		return false;
	},

	hasAttribute: function(item, attribute){
		 this._assertIsItem(item);
		 this._assertIsAttribute(attribute);
		 if (item[attribute]) {
			 return true;
		 }
		 return false;
	},

	getAttributes: function(item){
		 this._assertIsItem(item);
		 var attrs = [];
		 var i;
		 for (i in item) {
			 if (i !== this._storeRef) {
				 attrs.push(i);
			 }
		 }
		 return attrs;
	},

	getValue: function(item, attribute){
		 var values = this.getValues(item, attribute);
		 if (values) {
			 return values[0];
		 }
		 return undefined;
	},

	getValues: function(item, attribute){
		 this._assertIsItem(item);
		 this._assertIsAttribute(attribute);
		 var value = item[attribute];
		 if (value !== undefined) {
			 if (!dojo.isArray(value)) {
				 value = [value];
			 }
		 }
		 return value;
	},

	getLabel: function(item) {
		 this._assertIsItem(item);
		 return this.getValue(item, this.memberLabel);
	},

	getLabelAttributes: function(item) {
		 return [this.memberLabel];
	},

	containsValue: function(/* item */ item, 
							/* attribute-name-string */ attribute, 
							/* anything */ value){
		//	summary: 
		//		See dojo.data.api.Read.containsValue()
		var regexp = undefined;
		if(typeof value === "string"){
			regexp = dojo.data.util.filter.patternToRegExp(value, false);
		}
		return this._containsValue(item, attribute, value, regexp); //boolean.
	},

	isItemLoaded: function(/* anything */ something){
		//	summary: 
		//		See dojo.data.api.Read.isItemLoaded()
		return this.isItem(something); //boolean
	},

	loadItem: function(/* object */ keywordArgs){
		//	summary: 
		//		See dojo.data.api.Read.loadItem()
		this._assertIsItem(keywordArgs.item);
	},

	fetch: function(request){

		request = request || {};
		if(!request.store){
			request.store = this;
		}
		var self = this;
		var scope = request.scope || dojo.global;

		//Generate what will be sent over.
		var reqParams = {};

		var xhrData = {
			url: this.resourceUri,
			preventCache: true,
			handleAs: "json-comment-optional",
			content: reqParams,
			preventCache: true
		};

		var deferred = dojo.xhrGet(xhrData);
		deferred.addCallback(function(items){
			items = self._processItems(items);

            var total = items.length;
			//Sort it if we need to.
			if (request.sort) {
				items.sort(dojo.data.util.sorter.createSortFunction(request.sort, self));
			}

			//Serverside doesn't handle paging, so have to do that here.
			var start = 0;
			var count = items.length;
			if(request.start > 0 && request.start < items.length) {
				start = request.start;	
			}
			if (request.count && request.count) {
				count = request.count;
			}
			var endIdx = start + count;
			if (endIdx > items.length) {
				endIdx = items.length;
			}

			items = items.slice(start, endIdx);

			if (request.onBegin) {
				request.onBegin.call(scope, total, request);
			}
            if(request.onItem){
				if(dojo.isArray(items)){
					var i = 0;
					for (i = 0; i < items.length; i++) {
						request.onItem.call(scope, items[i], request); 
					}
					if (request.onComplete) {
						request.onComplete.call(scope, null, request);
					}
				}
			}else if (request.onComplete){
				request.onComplete.call(scope, items, request); 
			}
		});
		deferred.addErrback(function(error){
			if(request.onComplete){
				request.onError.call(scope, error, request);
			}
		});
		return request;
	},

	close: function() {
		 //Do nothing.
	},

	_processItems: function(itemsArray){
		var items = [];
		this._arrayOfItems = [];
		var i;
		for (i = 0; i < itemsArray.length; i++) {
			var item = itemsArray[i];
			item[this._storeRef] = this;
			this._arrayOfItems[item[this.memberIdentifier]] = item;
			items.push(item);
		}
		return items;
	},
	
	_assertIsItem: function(/* item */ item){
		//	summary:
		//      This function tests whether the item passed in is indeed an item in the store.
		//	item: 
		//		The item to test for being contained by the store.
		if(!this.isItem(item)){ 
			throw new Error("zero.resource.RestStore: Invalid item argument.");
		}
	},

	_assertIsAttribute: function(/* attribute-name-string */ attribute){
		//	summary:
		//		This function tests whether the item passed in is indeed a valid 'attribute' like type for the store.
		//	attribute: 
		//		The attribute to test for being contained by the store.
		if(typeof attribute !== "string"){ 
			throw new Error("zero.resource.RestStore: Invalid attribute argument.");
		}
	},

	_containsValue: function(	/* item */ item, 
								/* attribute-name-string */ attribute, 
								/* anything */ value,
								/* RegExp?*/ regexp){
		//	summary: 
		//		Internal function for looking at the values contained by the item.
		//	description: 
		//		Internal function for looking at the values contained by the item.  This 
		//		function allows for denoting if the comparison should be case sensitive for
		//		strings or not (for handling filtering cases where string case should not matter)
		//	
		//	item:
		//		The data item to examine for attribute values.
		//	attribute:
		//		The attribute to inspect.
		//	value:	
		//		The value to match.
		//	regexp:
		//		Optional regular expression generated off value if value was of string type to handle wildcarding.
		//		If present and attribute values are string, then it can be used for comparison instead of 'value'
		return dojo.some(this.getValues(item, attribute), function(possibleValue){
			if(possibleValue !== null && !dojo.isObject(possibleValue) && regexp){
				if(possibleValue.toString().match(regexp)){
					return true; // Boolean
				}
			}else if(value === possibleValue){
				return true; // Boolean
			}
		});
	},

	getIdentity: function(item) {
		 return item[this.memberIdentifier];
	},

	fetchItemByIdentity: function(keywordArgs) {
		var identity = keywordArgs.identity;
		var args = {
  			url: this.resourceUri + "/" + keywordArgs.identity,
			handleAs: "json-comment-optional",
			preventCache: true
		};
		
		var self = this;
		var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
		var deferred = dojo.xhrGet(args);
		deferred.addCallback(function(data) {
			var item = null;
			if (data !== null) {
				item = data;
				item[self._storeRef] = self;
			}
			if(keywordArgs.onItem){
				keywordArgs.onItem.call(scope, item);
			}
		});
		deferred.addErrback(function(error){
			if(keywordArgs.onError){
				keywordArgs.onError.call(scope, error);
			}
		});
	},

	getIdentityAttributes: function(item) {
		return [this.memberIdentifier];
	},

	newItem: function(keywordArgs, parentInfo) { 
		var data = null;
        var self = this;
        var err = null;
		var args = {
			url: this.resourceUri,
			handleAs: "json-comment-optional",
			postData: dojo.toJson(keywordArgs),
			preventCache: true,
			sync: true,
		    contentType: "application/json",
            handle: function(resp, ioArgs) {
                //IE returns 1223 on a 204 response.  Argh.
                if(ioArgs.xhr.status === 1223 || ioArgs.xhr.status === 204){
                    data = keywordArgs;
                    data[self._storeRef] = self;
                }else{
                    err = new Error(ioArgs.xhr.status + ": " + ioArgs.xhr.statusText);
                    err.responseText = ioArgs.xhr.responseText;
                }
            }
		};
		var deferred = dojo.rawXhrPost(args);
        if(err){
            throw err;
        }
        this.onNew(data,parentInfo);
		return data;
	},

	deleteItem: function(item) {
		this._assertIsItem(item);
		var args = {
			url: this.resourceUri + "/" + escape(this.getValue(item, this.memberIdentifier)),
			handleAs: "json-comment-optional",
			preventCache: true,
			sync: true
		};
		var deferred = dojo.xhrDelete(args);
		deferred.addErrback(function(error) {
			throw error;
		});
        this.onDelete(item);
	},

	setValue: function(item, attribute, value) {
        if(!this._changedItems){
            this._changedItems = {};
        }
		this._assertIsItem(item);
		this._assertIsAttribute(attribute);
        var oldValue = item[attribute];
		item[attribute] = value;
        this._changedItems[item[this.memberIdentifier]] = item;
        this.onSet(item,attribute,oldValue,value);
	},

	setValues: function(item, attribute, value) {
		this._assertIsItem(item);
		this._assertIsAttribute(attribute);
        if(!dojo.isArray(value)){
            value = [value];
        }
        var oldValue = item[attribute];
		item[attribute] = value;
        this._changedItems[item[this.memberIdentifier]] = item;
        this.onSet(item,attribute,oldValue,value);
	},

	unsetAttribute: function(item, attribute) {
		this._assertIsItem(item);
		this._assertIsAttribute(attribute);
		delete item[attribute];
        this._changedItems[item[this.memberIdentifier]] = item;
	},
    
	save: function(/* object */ keywordArgs) {
		 var scope = keywordArgs.scope || dojo.global;
		 var item = null, obj = null;
		 for(var id in this._changedItems){
		 	item = this._changedItems[id];
		 	obj = {};
		 	for(var key in item){
		 		if(key != this._storeRef){
		 			obj[key] = item[key];
		 		}
		 	}
		 	var args = {
		 		url: this.resourceUri + "/" + escape(this.getValue(item, this.memberIdentifier)),
		 		putData: dojo.toJson(obj),
		 		handleAs: "json-comment-optional",
		 		preventCache: true,
                contentType: "application/json",
			 	sync: true
		 	};
		 	var deferred = dojo.rawXhrPut(args);
		 }
		 if (keywordArgs.onComplete) {
			 keywordArgs.onComplete.call(scope);
             this._changedItems = {};
		 } else if (keywordArgs.onError) {
             keywordArgs.onError.call(scope);
         }
	},

	revert: function(){
		throw new Error("This store cannot revert!");
	},

    /* dojo.data.api.Notification */

	onSet: function(/* item */ item, 
					/*attribute-name-string*/ attribute, 
					/*object | array*/ oldValue,
					/*object | array*/ newValue){
		// summary: See dojo.data.api.Notification.onSet()
		
		// No need to do anything. This method is here just so that the 
		// client code can connect observers to it.
	},

	onNew: function(/* item */ newItem, /*object?*/ parentInfo){
		// summary: See dojo.data.api.Notification.onNew()
		
		// No need to do anything. This method is here just so that the 
		// client code can connect observers to it. 
	},

	onDelete: function(/* item */ deletedItem){
		// summary: See dojo.data.api.Notification.onDelete()
		
		// No need to do anything. This method is here just so that the 
		// client code can connect observers to it. 
	}
});